/*
 *  Copyright (c) 2021, 2022 Daniel Wilkins.
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.

 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 ****************************************************************************/

#include <stdarg.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include <unistd.h>

// Make sure we have access to the super key.
#define XK_MISCELLANY
#define XK_LATIN1
#include <X11/keysymdef.h>
#include <X11/Xlib.h>

#include <X11/XKBlib.h>
#include <X11/cursorfont.h>

#include "logos.h"

// Macros
#define max(a, b) (a > b ? a : b)
#define min(a, b) (a < b ? a : b)

#define DEFAULT_BG_COLOR 0x444499
// Types
enum WMMode {
    WM_NORMAL,
    WM_REDRAW,
    WM_REDRAW_DRAWING,
    WM_MENU
};

enum RadialType {
    MENU_WINDOW,
    MENU_ROOT,
    MENU_END
};


struct WMState {
    // The display
    Display *dpy;
    Window root;
    // The currently-focused window.
    Window focus;
    // What mode we're in (redrawing, normal, menu open, etc.)
    enum WMMode mode;

    // Which type of radial menu we're using (quicklaunch or window)
    enum RadialType radial_type;

    // Stash for the initial x/y of a redraw
    int redraw_start_x;
    int redraw_start_y;
    // Resources for the redrawing box
    Window redraw_box;

    // Last known mouse x and y coords
    int last_mouse_x;
    int last_mouse_y;

    // argv
    char **argv;
};

// Signature for a radial button callback
typedef void (*RadialCB)(struct WMState *wm);

// A radial menu entry
struct RadialEntry {
    char *label;
    RadialCB cb;
    GC gc;
    Window w;

};

enum RadialPosition {
    RAD_N,
    RAD_E,
    RAD_S,
    RAD_W,
    RAD_END
};

// A radial menu is just the 4 buttons.
// Organized in NESW
struct RadialMenu {
    int colors[RAD_END];    // Gather up all the already-existing windows and make the wm know about them.
    struct RadialEntry buttons[RAD_END];
};

// Just a return type for a utility function defining a rectangle
struct DrawRect {
    int x, y, w, h;
};

// Globals
// These are some X magic atoms. Don't worry about them too much.
static Atom WM_PROTOCOLS;
static Atom WM_DELETE_WINDOW;

// Convenience variable for the super key.
static unsigned int SUPER_KEY;

// String representations of the wm modes.
char *wm_modes[] = {"WM_NORMAL", "WM_REDRAW", "WM_REDRAW_REDRAWING", "WM_MENU"};

// Forward declaration of radial callbacks so we can define the menu here.
void radial_raise_cb(struct WMState *wm);
void radial_close_cb(struct WMState *wm);
void radial_sink_cb(struct WMState *wm);
void radial_redraw_cb(struct WMState *wm);
void radial_term_cb(struct WMState *wm);
void radial_browse_cb(struct WMState *wm);
void radial_run_cb(struct WMState *wm);
void radial_fm_cb(struct WMState *wm);

// Definition for the radial menus.
struct RadialMenu radial_menus[2] = {
    // MENU_WINDOW, for menus when a window is hovered over
    {
        //         cyan      magenta   yellow    grey
        .colors = {0x00aaaa, 0xaa00aa, 0xaaaa00, 0xaaaaaa},

        // The buttons themselves
        .buttons = {
            {.label = "Raise", .cb = radial_raise_cb},
            {.label = "Close", .cb = radial_close_cb},
            {.label = "Sink", .cb = radial_sink_cb},
            {.label = "Redraw", .cb = radial_redraw_cb}}},
    // MENU_ROOT, for when the root is selected
    {
        //         green     red       blue      white
        .colors = {0x00aa00, 0xaa0000, 0x0000aa, 0xdddddd},

        .buttons = {
            {.label = "Terminal", .cb = radial_term_cb},
            {.label = "Browser", .cb = radial_browse_cb},
            {.label = "Run", .cb = radial_run_cb},
            {.label = "FM", .cb = radial_fm_cb}}}};

// For the 2 cursors we use
enum Cursors {
    CURSOR_NORMAL,
    CURSOR_REDRAW
};

Cursor cursors[2] = {0};

struct WindowSize {
    int w, h;
}resizing_window;

// Utility functions

// Set focus to window w
void focus_window(struct WMState *wm, Window w) {
    wm->focus = w;

    XGrabKey(wm->dpy, SUPER_KEY, AnyModifier, w, true, GrabModeAsync, GrabModeAsync);
    XSetInputFocus(wm->dpy, w, RevertToParent, CurrentTime);

}

// Some display managers (SDDM for one) don't clean up the framebuffer when they launch a wm, leaving garbage on the screen.
// Set the background color of the root window to fix this.
void clear_root_window(struct WMState *wm) {
    XSetWindowBackground(wm->dpy, wm->root, DEFAULT_BG_COLOR);
    XClearWindow(wm->dpy, wm->root);
}

// Set up our pre-existing windows
void configure_preexisting_windows(struct WMState *wm) {
    // Gather up all the already-existing windows and make the wm know about them.
    Window *children;
    Window _root, _parent;
    unsigned int num_children;
    // Get the list of already-existing children of root
    XQueryTree(wm->dpy, wm->root, &_root, &_parent, &children, &num_children);

    // For all the children.
    for (unsigned int i = 0; i < num_children; i++) {
        logos_log(LOG_DEBUG, "Found existing child %d.", children[i]);
        // Set up our listeners.
        XSelectInput(wm->dpy, children[i], EnterWindowMask | KeyReleaseMask | PointerMotionMask);

        // Fix up focus.
        Window child;
        // All unused but we need valid pointers for them
        int _x, _y, _rx, _ry;
        unsigned int _mask;
        if (XQueryPointer(wm->dpy, children[i], &_root, &child, &_x, &_y, &_rx, &_ry, &_mask)) {
            printf("CHILD IS %lx QUERIED ON %lx\n", child, children[i]);
	    if (child != 0) {
                printf("Focusing on window %lx\n", children[i]);
                XSetInputFocus(wm->dpy, children[i], RevertToParent, CurrentTime);
	    }
        }
    }

    // Free the list X created for us.
    if (children) {
        XFree(children);
    }
}

// Given 2 points, construct the rectangle between them.
struct DrawRect construct_draw_rect(int ox, int oy, int dx, int dy) {
    return (struct DrawRect) {
        .x = ox < dx ? ox : dx,
        .y = oy < dy ? oy : dy,
        .w = abs(dx - ox),
        .h = abs(dy - oy)
    };
}

bool is_point_outside_rect(int px, int py, int rx, int ry, int rw, int rh) {
    return (px < rx || px > (rx + rw)) || (py < ry || py > (ry + rh));
}


// Spawn a child
void spawn(Display *dpy, char **command) {
    // Create a new process
    if (fork() == 0) {
        logos_log(LOG_INFO, "Spawning %s", command[0]);
        // Close the connection to dpy so it doesn't cause problems for the child
        close(ConnectionNumber(dpy));
        // Create a new session so it doesn't die if the wm crashes
        setsid();
        // Execute the command
        execvp(command[0], command);

        // Should never be reached.
        logos_log(LOG_ERROR, "Failed to execute command %s", command[0]);
        exit(0);
    }
}

// Radial button helpers
void create_radial_windows(struct WMState *wm) {
    enum RadialType type = wm->radial_type;
    // For each of the buttons
    for (int j = 0; j < RAD_END; j++) {
        // Make a dummy 64x64 window
        radial_menus[type].buttons[j].w = XCreateSimpleWindow(wm->dpy, wm->root,
                                                           0, 0, 64, 64,
                                                           0, 0, radial_menus[type].colors[j]);
        // Create a graphics context for it so we can draw on it.
        radial_menus[type].buttons[j].gc = XCreateGC(wm->dpy, radial_menus[type].buttons[j].w, 0, 0);

        // Set the foreground color to black
        XSetForeground(wm->dpy, radial_menus[type].buttons[j].gc, 0x000000);

        // We want to hear key up events on these windows.
        XSelectInput(wm->dpy, radial_menus[type].buttons[j].w, KeyReleaseMask);
    }
}

// Free all of the radial menu entries when we're done.
void free_radials(struct WMState *wm) {
    enum RadialType type = wm->radial_type;
    for (int j = 0; j < RAD_END; j++) {
        XFreeGC(wm->dpy, radial_menus[type].buttons[j].gc);
        XDestroyWindow(wm->dpy, radial_menus[type].buttons[j].w);
    }
}

bool is_radial_button(Window w) {
    // For both of the radial menus
    for (int i = 0; i < RAD_END; i++) {
        // For every entry
        for (int j = 0; j < RAD_END; j++) {
            // If the window we're checking is in those entries, return true.
            if (w == radial_menus[i].buttons[j].w) {
                return true;
            }
        }
    }
    return false;
}

// Return the radial button entry for a window.
struct RadialEntry *radial_button(Window w) {
        for (int i = 0; i < MENU_END; i++) {
            for (int j = 0; j < RAD_END; j++) {
                if (w == radial_menus[i].buttons[j].w) {
                    return &(radial_menus[i].buttons[j]);
                }
            }
        }
    // Should never return null
    // (because I only call it after checking is_radial_button
    return NULL;
}

// Move the radicals to the proper locations around the cursor, possibly warping
// the cursor if it's too close to an edge.
void warp_radials(struct WMState *wm, long cx, long cy) {
    struct RadialMenu menu = radial_menus[wm->radial_type];
    // Make sure the boxes are on-screen
    int screen = DefaultScreen(wm->dpy);
    int width = DisplayWidth(wm->dpy, screen);
    int height = DisplayHeight(wm->dpy, screen);

    // 96 because 64 pixels for a window to the left + 32 padding
    int px = min(width - 96, max(96, cx));
    int py = min(height - 96, max(96, cy));

    // We assume that the screen is at least 192x192. That seems fairly reasonable.
    logos_log(LOG_DEBUG, "Warping pointer to %d %d", px, py);
    XWarpPointer(wm->dpy, None, wm->root,
                 0, 0, 0, 0,
                 px, py);

    XMoveWindow(wm->dpy, menu.buttons[0].w, px - 32, py - 96);
    XMoveWindow(wm->dpy, menu.buttons[1].w, px + 32, py - 32);
    XMoveWindow(wm->dpy, menu.buttons[2].w, px - 32, py + 32);
    XMoveWindow(wm->dpy, menu.buttons[3].w, px - 96, py - 32);

    // Bring them all to the front and draw their text naively
    for (int i = 0; i < RAD_END; i++) {
        XRaiseWindow(wm->dpy, menu.buttons[i].w);
        XMapWindow(wm->dpy, menu.buttons[i].w);
        XDrawString(wm->dpy, menu.buttons[i].w, menu.buttons[i].gc, 16, 32, menu.buttons[i].label, strlen(menu.buttons[i].label));
    }
    XSync(wm->dpy, false);
}

// Radial menu callbacks

// Window menu options

// Close the focused window
// This can technically fail to close poorly behaved windows, but that's
// what xkill is for.
void radial_close_cb(struct WMState *wm) {
    logos_log(LOG_DEBUG, "Closing focused window nicely.");

    // Populate the message sent to the window being closed.
    XEvent message = {
        .xclient = {
            .type = ClientMessage,
            .message_type = WM_PROTOCOLS,
            .window = wm->focus,
            .format = 32,
            .data = {
                .l = {WM_DELETE_WINDOW}
            }
        }
    };
    XSendEvent(wm->dpy, wm->focus, false, 0, &message);
}

// Send the focused window to the top of the stack
void radial_raise_cb(struct WMState *wm) {
    XRaiseWindow(wm->dpy, wm->focus);
}

// Redraw the focused window.
// We need user input for the rest of this, just put us in the relevant state
void radial_redraw_cb(struct WMState *wm) {
    logos_log(LOG_DEBUG, "Redraw of window %d requested", wm->focus);
    wm->mode = WM_REDRAW;
    XWindowAttributes window_info;
    XGetWindowAttributes(wm->dpy, wm->focus, &window_info);

    resizing_window.w = window_info.width;
    resizing_window.h = window_info.height;
    // Set the cursor to the redraw one
    XDefineCursor(wm->dpy, wm->root, cursors[CURSOR_REDRAW]);
    // Give all focus to the root window
    XSetInputFocus(wm->dpy, wm->root, RevertToParent, CurrentTime);
    // Hide the window we're redrawing.
    XUnmapWindow(wm->dpy, wm->focus);
    // Initialize the box we use to show windows we're redrawing
    wm->redraw_box = XCreateSimpleWindow(wm->dpy, wm->root,
                                                 // Rect
                                                 0, 0, 1, 1,
                                                 // Border width
                                                 4,
                                                 // Border and background colors
                                                 // Cyan and grey
                                                 0x00ffff, 0xbbbbbb);

    XGrabPointer(wm->dpy, wm->root, false, PointerMotionMask, GrabModeAsync, GrabModeAsync, None, cursors[CURSOR_REDRAW], CurrentTime);

}

// Send the focused window to the bottom of the stack
void radial_sink_cb(struct WMState *wm) {
    XLowerWindow(wm->dpy, wm->focus);
}


// Root/quicklaunch menu options

// Launch a terminal
void radial_term_cb(struct WMState *wm) {
    char *cmd[] = {"urxvt", NULL};
    spawn(wm->dpy, cmd);
}

// Launch a web browser
void radial_browse_cb(struct WMState *wm) {
    char *cmd[] = {"firefox", NULL};
    spawn(wm->dpy, cmd);
}

// Launch an arbitrary program
void radial_run_cb(struct WMState *wm) {
    char *cmd[] = {"gmrun", NULL};
    spawn(wm->dpy, cmd);
}

// Launch a file manager
void radial_fm_cb(struct WMState *wm) {
    char *cmd[] = {"xfe", NULL};
    spawn(wm->dpy, cmd);
}


// Window Manager Event Callbacks

void configure_cb(struct WMState *wm, XConfigureRequestEvent event) {
    // The struct we have to pass on to X to actually make our changes.
    // We mostly just copy it from the window's request.
    XWindowChanges changes = {
        .x = event.x,
        .y = event.y,
        .width = event.width,
        .height = event.height,
        .border_width = event.border_width,
        .sibling = event.above,
        .stack_mode = event.detail};

    // Now we actually act on the request
    XConfigureWindow(wm->dpy, event.window, event.value_mask, &changes);
    XSync(wm->dpy, false);
    logos_log(LOG_DEBUG, "Resizing window %ld to %dx%d\n", event.window, event.width,
              event.height);

}

//Called when a window is created (before it's shown)
void create_cb(struct WMState *wm, XCreateWindowEvent event) {
    XWindowAttributes window_info;
    XGetWindowAttributes(wm->dpy, event.window, &window_info);

    if (event.window != wm->redraw_box && !window_info.override_redirect) {
        // Listen to the super key on the window.
        XGrabKey(wm->dpy, SUPER_KEY, AnyModifier, event.window, true, GrabModeAsync, GrabModeAsync);
        if (!is_radial_button(event.window)) {
            XSelectInput(wm->dpy, event.window, EnterWindowMask | KeyReleaseMask | PointerMotionMask);
            // Initially move the window under the mouse cursor.
            XMoveWindow(wm->dpy, event.window, max(0, wm->last_mouse_x - (event.width / 2)), max(0, wm->last_mouse_y - (event.height / 2)));
        }
    }

}

// Just log errors as we get them.
// They'll only be useful to developers.
int error_cb(Display *dpy, XErrorEvent *err) {
    logos_log(LOG_DEBUG, "Xlib Error! code %d: request: %d.%d\n", err->error_code,
      err->request_code, err->minor_code);
    return 0;
}

// Called when the mouse enters a window
void focus_cb(struct WMState *wm, XCrossingEvent event) {
    logos_log(LOG_DEBUG, "Received focus event");
    // If we're in a mode which cares about window entries
    if (event.window != event.root
        && wm->mode == WM_NORMAL) {
        logos_log(LOG_DEBUG, "Received focusing on non-root. New focus window: %d", event.window);
        // Set the new window as our focus
        focus_window(wm, event.window);
    }
}


void key_down_cb(struct WMState *wm, XKeyEvent event) {
    logos_log(LOG_DEBUG, "Received down key event. Window? %d Focused %d In radial? %d State? %x Code? %x", event.window, wm->focus, is_radial_button(event.window), event.state, event.keycode);
    logos_log(LOG_DEBUG, "Window mode %s", wm_modes[wm->mode]);
    // A place to store the attributes of the focused window.
    XWindowAttributes window_info;

    // See if it's the menu key
    if (event.keycode == SUPER_KEY) {
        // Ignore repeats
        if (wm->mode == WM_NORMAL) {

            // Make sure it wasn't caused by the window just because of focus.
            // If the pointer is on the root window, it should spawn the root menu even if a window
            // has focus
            XGetWindowAttributes(wm->dpy, event.window, &window_info);

            // If the super press was in the root window, even if the keyboard event "came from" another
            // window (because of sloppy focus,) run the root radial menu.
            if (event.window == wm->root
                || is_point_outside_rect(event.x_root, event.y_root,
                                         window_info.x, window_info.y,
                                         window_info.width, window_info.height)) {
                logos_log(LOG_DEBUG, "Triggering a root menu.");

                wm->radial_type = MENU_ROOT;
            } else {
                // If it's genuinely in a window, go for that.
                logos_log(LOG_DEBUG, "Triggering a window menu. Ungrabbing super on %d", event.window);

                // Let go of the super key so that we can catch it in the radial buttons.
                XUngrabKey(wm->dpy, SUPER_KEY, AnyModifier, event.window);
                // We're now running a radial menu.
                wm->radial_type = MENU_WINDOW;
            }

            logos_log(LOG_DEBUG, "Creating radial windows");
            // Create and initialize our radial menus
            create_radial_windows(wm);

            // Put the radial menu in its place.
            warp_radials(wm, event.x_root, event.y_root);

            // Give focus to root
            XSetInputFocus(wm->dpy, wm->root, RevertToParent, CurrentTime);

            // We're now in menu mode.
            logos_log(LOG_DEBUG, "Setting wm mode to MENU");
            wm->mode = WM_MENU;
        } else if (wm->mode == WM_REDRAW) {
            logos_log(LOG_DEBUG, "Beginning redrawing.");
            // Store the coordinates of the window-to-be's corner
            wm->redraw_start_x = event.x_root;
            wm->redraw_start_y = event.y_root;

            // Put us into active redraw mode.
            logos_log(LOG_DEBUG, "Setting wm mode to REDRAWING_DRAWING");
            wm->mode = WM_REDRAW_DRAWING;
        }
    } else if (event.keycode == XKeysymToKeycode(wm->dpy, XK_R) && event.state & Mod4Mask) {
        logos_log(LOG_ERROR, "RELAUNCHING WITH %s", wm->argv[0]);
        execvp(wm->argv[0], wm->argv);
    } else if (event.keycode == XKeysymToKeycode(wm->dpy, XK_Q) && event.state & Mod4Mask) {
        logos_log(LOG_ERROR, "Quitting");
        exit(0);
    }

}

void key_up_cb(struct WMState *wm, XKeyEvent event) {
    logos_log(LOG_DEBUG, "Received up key event. Window? %d Focused %d In radial? %d State? %x Code? %x", event.window, wm->focus, is_radial_button(event.window), event.state, event.keycode);

    if (event.keycode == SUPER_KEY) {
        // If the super key was down because we had a menu open...
        if (wm->mode == WM_MENU) {
            // Check if we clicked a radial button
            if (is_radial_button(event.window)) {
                struct RadialEntry *button = radial_button(event.window);
                logos_log(LOG_INFO, "Radial button %s pressed", button->label);

                // If it was, we do its action.
                button->cb(wm);
            }
            // Make the radial buttons disappear
            for (int i = 0; i < RAD_END; i++) {
                XUnmapWindow(wm->dpy, radial_menus[wm->radial_type].buttons[i].w);
            }
            free_radials(wm);
            logos_log(LOG_INFO, "Re-grabbing super on window %d (Don't worry bout an error right after this)", wm->focus);

            if (wm->mode != WM_REDRAW) {
                // Grab Super on the focused window again.
                XGrabKey(wm->dpy, SUPER_KEY, AnyModifier, wm->focus, true, GrabModeAsync, GrabModeAsync);
                wm->mode = WM_NORMAL;
                logos_log(LOG_DEBUG, "Setting wm mode to NORMAL");
            }


        // If we released super because we were redrawing a window.
        } else if (wm->mode == WM_REDRAW_DRAWING) {
            // Construct the normalized rectangle for the draw.
            struct DrawRect dr = construct_draw_rect(wm->redraw_start_x, wm->redraw_start_y,
                                                     event.x_root, event.y_root);

            // A place to store the attributes of the focused window.
            XWindowAttributes window_info;
            XGetWindowAttributes(wm->dpy, wm->focus, &window_info);

            logos_log(LOG_INFO, "Original window is %d by %d", window_info.width, window_info.height);
            bool moved = false;

            if (dr.w <= 16 ||  dr.h <= 16) {
                moved = true;
            }

            // If the user didn't draw a rect for us, move the window without resizing.
            // Otherwise use some sensible defaults in case the user drew an unreasonably tiny window.
            int new_width = dr.w == 0 ? window_info.width : max(dr.w, min(256, window_info.width));
            int new_height = dr.h == 0 ? window_info.height : max(dr.h, min(256, window_info.height));
            // Move our window to that location/size
            XWindowChanges c = {
            .x = moved ? dr.x - (new_width / 2) : dr.x,
            .y = moved ? dr.y - (new_height / 2) : dr.y,
            .width = new_width,
            .height = new_height};
            // Commit the changes
            XConfigureWindow(wm->dpy, wm->focus, CWX | CWY | CWWidth | CWHeight, &c);
            // Hide the redraw box
            XUnmapWindow(wm->dpy, wm->redraw_box);
            // Show the newly moved/resized window.
            XMapWindow(wm->dpy, wm->focus);
            // Set the cursor back to normal.
            XDefineCursor(wm->dpy, wm->root, cursors[CURSOR_NORMAL]);
            // Put focus on the window if we just did a move, otherwise put it on the root.
            XSetInputFocus(wm->dpy, moved ? wm->focus : wm->root, RevertToParent, CurrentTime);

            XSync(wm->dpy, false);

            // Put the WM back into its normal state.
            logos_log(LOG_DEBUG, "Setting wm mode to NORMAL");

            XDestroyWindow(wm->dpy, wm->redraw_box);
            XUngrabPointer(wm->dpy, CurrentTime);
            wm->mode = WM_NORMAL;
        }
    }

}

void map_cb(struct WMState *wm, XMapRequestEvent event) {
    XMapWindow(wm->dpy, event.window);
     Window *_children;
    Window _root, parent;
    unsigned int _num_children;
    // Get the list of already-existing children of root
    XQueryTree(wm->dpy, event.window, &_root, &parent, &_children, &_num_children);
    if (!is_radial_button(event.window)
        && event.window != wm->redraw_box
        && parent == wm->root
        && wm->mode == WM_NORMAL) {
        focus_window(wm, event.window);
    }
}

void pointer_motion_cb(struct WMState *wm, XMotionEvent event) {
    wm->last_mouse_x = event.x_root;
    wm->last_mouse_y = event.y_root;
    // We only care about mouse motion if we're redrawing a window.
    if (wm->mode == WM_REDRAW_DRAWING) {
        // Convert to a DrawRect, which will be correct whether which way the user swept.
        struct DrawRect dr = construct_draw_rect(wm->redraw_start_x, wm->redraw_start_y,
                                                 event.x_root, event.y_root);
        XWindowChanges c = {
            .x = dr.x,
            .y = dr.y,
            .width = dr.w,
            .height = dr.h};

        // Move the window to its new position/size
        XConfigureWindow(wm->dpy, wm->redraw_box, CWX | CWY | CWWidth | CWHeight, &c);
        // Make sure it appears over all other windows.
        XRaiseWindow(wm->dpy, wm->redraw_box);

        XSync(wm->dpy, false);

        // Make the redraw ghost box appear
        XMapWindow(wm->dpy, wm->redraw_box);
    } else if (wm->mode == WM_REDRAW) {
        int hw = resizing_window.w / 2;
        int hh = resizing_window.h / 2;
        XWindowChanges c = {
            .x = event.x_root - hw,
            .y = event.y_root - hh,
            .width = resizing_window.w,
            .height = resizing_window.h};

        printf("Configure!\n");
        // Move the window to its new position/size
        XConfigureWindow(wm->dpy, wm->redraw_box, CWX | CWY | CWWidth | CWHeight, &c);
        XSync(wm->dpy, false);
        printf("Raise!\n");
        // Make sure it appears over all other windows.
        XRaiseWindow(wm->dpy, wm->redraw_box);

        XSync(wm->dpy, false);

        printf("Map!\n");
        // Make the redraw ghost box appear
        XMapWindow(wm->dpy, wm->redraw_box);
    }
}

// This is only called if there's already a window manager running.
// It's torn down early in startup
int wm_running_cb(Display *dpy, XErrorEvent *err) {
    // BadAccess == WM is already running
    if (err->error_code != BadAccess) {
        logos_log(LOG_ERROR, "Something extra weird happened. Unexpected error code: %d\n",
                err->error_code);
        exit(2);
    }

    logos_log(LOG_ERROR, "Unable to grab root window. Is another wm running?");
    exit(0);

    // Satisfy the compiler
    return 0;
}

// Main

// Do all of the required setup before we can manage windows.
struct WMState wm_init(char **argv) {
    struct WMState ret = {
        .dpy = XOpenDisplay(NULL)
    };

    ret.argv = argv;

    if (ret.dpy == NULL) {
        logos_log(LOG_ERROR, "Failed to open display %s\n", XDisplayName(NULL));
        exit(1);
    }

    // Store the root window
    ret.root = DefaultRootWindow(ret.dpy);

    // Before we try to grab the root window, set up a handler in case
    // there's another WM running
    // (as seen by us being unable to grab the root window)
    XSetErrorHandler(&wm_running_cb);

    // Accept events on the root window.
    // Substructure events let us know when a child of root is doing something.
    // Button and Key masks let us listen to mouse and keyboard events
    // EnterWindow lets us listen to when focus should change
    // PointerMotion is just used for seeing when we're drawing a window basically
    XSelectInput(ret.dpy, ret.root, SubstructureRedirectMask | SubstructureNotifyMask
                 | ButtonPressMask | KeyPressMask | KeyReleaseMask
                 | EnterWindowMask | PointerMotionMask);

    cursors[CURSOR_NORMAL] = XCreateFontCursor(ret.dpy, XC_left_ptr);
    cursors[CURSOR_REDRAW] = XCreateFontCursor(ret.dpy, XC_crosshair);

    // Set the cursor to a reasonable default on the root window
    XDefineCursor(ret.dpy, ret.root, cursors[CURSOR_NORMAL]);

    // Force a protocol sync and fire the error callback if needed.
    XSync(ret.dpy, false);

    // Successfully initialized, use our usual error handler.
    XSetErrorHandler(&error_cb);

    clear_root_window(&ret);
    configure_preexisting_windows(&ret);

    // Grab the magic atoms needed to shut down windows
    WM_PROTOCOLS = XInternAtom(ret.dpy, "WM_PROTOCOLS", false);
    WM_DELETE_WINDOW = XInternAtom(ret.dpy, "WM_DELETE_WINDOW", false);

    // Store the keycode for the super key
    SUPER_KEY = XKeysymToKeycode(ret.dpy, XK_Super_L);

    return ret;
}

int main(int argc, char *argv[]) {
    // Initialize the logging library
    logos_init(NULL, LOG_DEBUG);

    logos_log(LOG_DEBUG, "argv[0] is %s", argv[0]);
    struct WMState wm = wm_init(argv);
    for (;;) {
        XEvent e;
        bool fake_event = false;
        // Pull the latest X event
        XNextEvent(wm.dpy, &e);

        switch (e.type) {
        // Birth of a new window
        case CreateNotify:
            create_cb(&wm, e.xcreatewindow);
            break;
        // Window configuration (initial location/size)
        case ConfigureRequest:
            configure_cb(&wm, e.xconfigurerequest);
            break;

        // Window wants to become visible
        case MapRequest:
            map_cb(&wm, e.xmaprequest);
            break;
        // Key pressed
        case KeyPress:
            key_down_cb(&wm, e.xkey);
            break;
        // Key pressed
        case KeyRelease:
            // A bunch of nonsense to deal with fake key up events.
            // There are theoretically X functions for this but they don't
            // help if your hardware is doing it.
            // First see if we have an event in the queue immediately after
            if (XEventsQueued(wm.dpy, QueuedAfterReading)) {
                XEvent pe;
                // If there is one, see what it is.
                XPeekEvent(wm.dpy, &pe);
                // If it's a key press identical to our key up at the same time, then
                // it's fake. Otherwise it's genuine
                if (pe.type == KeyPress
                    && pe.xkey.time == e.xkey.time
                    && pe.xkey.keycode == e.xkey.keycode) {
                    fake_event = true;
                }
            }
            if (!fake_event) {
                key_up_cb(&wm, e.xkey);
            }
            break;

        // Cursor enters a window
        case EnterNotify:
            if (wm.mode == WM_NORMAL) {
                focus_cb(&wm, e.xcrossing);
            }
            break;
        // Cursor moves
        case MotionNotify:
            pointer_motion_cb(&wm, e.xmotion);
            break;
        default:
            break;
        }
    }

    // Clean up the resources we allocated.
    XFreeCursor(wm.dpy, cursors[CURSOR_NORMAL]);
    XFreeCursor(wm.dpy, cursors[CURSOR_REDRAW]);

    free_radials(&wm);

    XCloseDisplay(wm.dpy);
}

